require "net/ldap"

module Devise

  module LDAP

    DEFAULT_GROUP_UNIQUE_MEMBER_LIST_KEY = 'uniqueMember'

    # Establishes connections and interacts with LDAP. Uses Connection objects to do this.    
    # Can interact with these methods in rails
    module Adapter

      # Returns all attributes for an account from LDAP
      def self.get_ldap_entry(login, scope = default_scope)
        self.ldap_connect(login, scope).search_for_login
      end

      # Get the value of an attribute for an account from LDAP
      def self.get_ldap_param(login, param, scope = default_scope)
        resource = self.ldap_connect(login, scope)
        resource.ldap_param_value(param)
      end

      # Returns the DistinguishedName of an account (regardless of if it exists or not)
      def self.get_dn(login, scope = default_scope)
        self.ldap_connect(login, scope).dn
      end

      def self.password_updatable? (login, scope = default_scope)
        options = { login: login, scope: scope }
        resource = Devise::LDAP::Connection.new(options)
        resource.password_updatable?
      end

      def self.user_creatable? (login, scope = default_scope)
        options = { login: login, scope: scope }
        resource = Devise::LDAP::Connection.new(options)
        resource.user_creatable?
      end

      # Boolean returned for if an account in the LDAP exists (doesn't check authentication / authorization): false if a valid match can't be obtained from ldap.
      def self.valid_login?(login, scope = default_scope)
        self.ldap_connect(login, scope).valid_login?
      end

      # Tries to authenticate credentails to LDAP. Returns true or false appropriately.
      def self.valid_credentials?(login, password_plaintext, scope = default_scope)
        options = { login: login, password: password_plaintext, scope: scope }
        resource = Devise::LDAP::Connection.new(options)
        resource.authorized?
      end

      # Returns true or false depending on if the users credentials have expired or not
      def self.expired_valid_credentials?(login, password_plaintext, scope = default_scope)
        options = { login: login, password: password_plaintext, scope: scope }
        resource = Devise::LDAP::Connection.new(options)
        resource.expired_valid_credentials?
      end

      # Updates a users password in LDAP
      def self.update_password(login, new_password, scope = default_scope)
        options = { login: login, new_password: new_password, scope: scope }
        resource = Devise::LDAP::Connection.new(options)
        resource.change_password! if new_password.present?
      end

      # Also updates the password. Unsure what differentiates this from update_password currently.
      def self.update_own_password(login, new_password, current_password)
        set_ldap_param(login, :userPassword, new_password, current_password, true)
      end

      # Returns a list of group memberships for a user
      def self.get_groups(login, scope = default_scope)
        self.ldap_connect(login, scope).user_groups
      end

      # Checks if a user is a member of a specific group
      def self.in_ldap_group?(login, group_name, group_attribute = nil, scope = default_scope)
        self.ldap_connect(login, scope).in_group?(group_name, group_attribute)
      end

      # Sets an LDAP attribute for an account to a new value
      def self.set_ldap_param(login, param, new_value, password = nil, scope = default_scope)
        options = { login: login, password: password, scope: scope }
        resource = Devise::LDAP::Connection.new(options)
        resource.set_param(param, new_value)
      end

      # Deletes the LDAP attribute for an account
      def self.delete_ldap_param(login, param, password = nil, scope = default_scope)
        options = { login: login, password: password, scope: scope }
        resource = Devise::LDAP::Connection.new(options)
        resource.delete_param(param)
      end

      # Creates a new connection to an LDAP database and returns the connection object to run methods against
      def self.ldap_connect(login, scope = default_scope)
        options = { login: login, scope: scope }
        Devise::LDAP::Connection.new(options)
      end

      # The default scope to use if none is specified
      def self.default_scope
        ::Devise.ldap_default_scope
      end

    end # module Adapter

  end # module LDAP

end # module Devise
